import os
import joblib
from typing import Union,Type
from abc import abstractmethod
import geoalchemy2 as ga
import sqlalchemy as sa
from sqlalchemy import Column,TIMESTAMP,Integer,Text,ARRAY,sql,REAL,ForeignKey,VARCHAR
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship,backref
from sqlalchemy.schema import UniqueConstraint
from ..global_support import reprer
from .statuses import image_status
from .utils import get_info
from ..folder_utils import Folder
from .. import log_support as logsup

Base=declarative_base()

class InfoBase(object):
	
	@property
	def info(self):
		return get_info(self)
	
	def __repr__(self):
		return reprer(self)

# def info(self):
# 	return mgr.get_wrapped(self)

class StatusTrackable(InfoBase):
	status_key='status'

class Reconnectable(InfoBase):
	
	@abstractmethod
	def reconnect(self,session):
		pass
	
	@staticmethod
	def get_dump_filepath(folder=None,filename='',filepath=''):
		if not filepath:
			assert all([folder,filename]),'Dump folder and filename should be provided.'
			filepath=Folder(folder).get_filepath(filename)
			if not os.path.splitext(filepath)[-1]: filepath+='.joblib'
		logsup.logger.debug('filepath: %s',filepath)
		return filepath
	
	@classmethod
	def from_dump(cls,folder=None,filename='',filepath=''):
		filepath=cls.get_dump_filepath(folder=folder,filename=filename,filepath=filepath)
		assert os.path.exists(filepath),f'Filepath {filepath} does not exist.'
		logsup.logger.debug('Loading object from dump file.')
		return joblib.load(filepath)
	
	def to_dump(self,folder=None,filename='',filepath=''):
		filepath=self.get_dump_filepath(folder=folder,filename=filename,filepath=filepath)
		joblib.dump(self,filepath)
		return self
	
	

class KeysBase(Reconnectable):
	primary_keys=['id']
	
	def reconnect(self,session):
		pass
	
	def __repr__(self):
		return reprer(self)

class IntegerID_Base(KeysBase):
	id=Column(Integer,primary_key=True)

class TextID_Base(KeysBase):
	id=Column(Text,primary_key=True)

class SnowTest(Base,IntegerID_Base):
	__tablename__='snow_test'
	geom=Column(ga.Geometry('MULTIPOLYGON'))

class Cadastres_Info(Base,IntegerID_Base):
	__tablename__='cadastres_cadastre'
	kad_nomer=Column(Text)
	area=Column(REAL)
	geom=Column(ga.Geometry('MULTIPOLYGON',srid=3857))

class Test_Srid(Base,IntegerID_Base):
	__tablename__='test_srid'
	ball_b=Column(REAL)
	descr=Column(Text)
	legend=Column(Text)
	geometry=Column(ga.Geometry(srid=3857))
	kad_nomer=Column(VARCHAR(20))
	area=Column(REAL)

class Test_Vectors(Base,IntegerID_Base):
	__tablename__='test_vectors'
	geometry=Column(ga.Geometry(srid=3857))
	index_value=Column(REAL)

# class Cadastres_Info2(Base,IntegerID_Base):
# 	__tablename__='cadastres_cadastre2'
# 	kad_nomer=Column(Text)
# 	area=Column(REAL)
# 	geom=Column(ga.Geometry('MULTIPOLYGON'))

class MainRequest(Base,IntegerID_Base,StatusTrackable):
	status_key='back_status'
	__tablename__='main_requests_request'
	request_date=Column(TIMESTAMP(True),server_default=sql.func.now())
	image_date=Column(TIMESTAMP(True),server_default=None)
	user_id=Column(Integer)
	order_request_date=Column(sa.Date,server_default=None)
	cadastre_id=Column(Integer)
	cadastre_value=Column(ga.Geometry('MULTIPOLYGON'))
	status=Column(Integer)
	status_changer=Column(Integer,default=0)
	back_status=Column(Text,default=image_status.Unprocessed)
	last_lookup_date=Column(TIMESTAMP(True),server_default=None)
	product_ids=Column(ARRAY(Text),server_default=None)
	results_dir=Column(Text,server_default=None)
	priority=Column(Integer,default=5)

class ImageDateChecker(Base,IntegerID_Base,StatusTrackable):
	status_key='image_date_checker_status'
	__tablename__='main_requests_imagedate'
	image_date=Column(TIMESTAMP(timezone=True))
	request_id=Column(Integer,ForeignKey(f'{MainRequest.__tablename__}.id'))
	cadastre_value=Column(ga.Geometry('MULTIPOLYGON'))
	image_date_checker_status=Column(Text)
	priority=Column(Integer,default=5)
	
	backend_request=relationship(MainRequest,uselist=False,backref=backref('image_date_request',cascade='all,delete'))

class User(Base,IntegerID_Base):
	__tablename__='auth_user'
	password=Column(sa.VARCHAR(128))
	last_login=Column(TIMESTAMP(timezone=True))
	is_superuser=Column(sa.Boolean,server_default='f')
	username=Column(sa.VARCHAR(150))
	first_name=Column(sa.VARCHAR(30))
	last_name=Column(sa.VARCHAR(150))
	email=Column(sa.VARCHAR(254))
	is_staff=Column(sa.Boolean,server_default='f')
	is_active=Column(sa.Boolean,server_default='t')
	date_joined=Column(TIMESTAMP(timezone=True))

class MeteoData(Base,KeysBase):
	__tablename__='meteo_data'
	primary_keys=['geometry','forecast_date']
	geometry=Column(ga.Geometry('POINT'),primary_key=True)
	forecast_date=Column(TIMESTAMP(timezone=True),primary_key=True)
	__table_args__=(UniqueConstraint('geometry','forecast_date',name='Unique_date_and_location'),)

class Sentinel2_Info(TextID_Base,Base,StatusTrackable):
	__tablename__='sentinel2_info'
	image_date=Column(TIMESTAMP(timezone=True))
	download_date=Column(TIMESTAMP(timezone=True))
	product_path_l1c=Column(Text,server_default=None)
	zip_path=Column(Text)
	product_path_l2a=Column(Text,server_default=None)
	status=Column(Text,server_default=image_status.Unprocessed)
	priority=Column(Integer,default=5)

class CeleryInfo(Base,IntegerID_Base):
	__tablename__='celery_taskmeta'
	task_id=Column(VARCHAR(155))
	status=Column(VARCHAR(50))
	date_done=Column(TIMESTAMP(True))
	traceback=Column(Text)
	
	def __repr__(self):
		return reprer(self)

class CadastreUser(Base,IntegerID_Base):
	__tablename__='cadastres_cadastreuser'
	
	title=Column(Text)
	cadastre_id=Column(Integer)
	user_id=Column(Integer)

def main():
	# Sentinel2_Info.__table__.create(engine)
	pass
	# old_statuses=collections.OrderedDict()
	# old_statuses['RunningAtmosphericCorrection']=image_status.Running.AtmosphericCorrection
	# old_statuses['FinishedAtmosphericCorrection']=image_status.Finished.AtmosphericCorrection
	# old_statuses['QueuedForAtmosphericCorrection']=image_status.QueuedFor.AtmosphericCorrection
	# old_statuses['NotForProcessing']=image_status.NotForProcessing
	#
	# for old_status,new_status in old_statuses.items():
	# 	for row in session.query(Sentinel2_Info).filter(Sentinel2_Info.status==old_status).all():
	# 		row.status=new_status
	pass
	
	pass

if __name__=='__main__':
	main()



import time
import pandas as pd
import lgblkb_tools.log_support as logsup
import multiprocessing as mp
import types
from typing import List
import uuid
import lgblkb_tools.scheduling.dbader_schedule as schedule


# region scheduled_task:
class ConveyorPosition:
	
	def __init__(self,job_func,max_process_count=1):
		self.job_func=job_func
		self.max_process_count=max_process_count
		self.current_process_count=0
		self.processes=list()
	
	def check_for_completed_processes(self):
		for p in self.processes:
			if not p.is_alive():
				p.join()
				p.terminate()
				self.processes.remove(p)
				self.current_process_count-=1
	
	def start_process(self,target,*args,name=None,**kwargs):
		if self.current_process_count<self.max_process_count:
			# simple_logger.info('Starting process...')
			p=mp.Process(target=target,args=args,kwargs=kwargs,name=name)
			p.start()
			self.processes.append(p)
			self.current_process_count+=1

def queue_generator(queue: mp.Queue,func,control_desk,output=None):
	if output is None:
		resultant_status,data=func()
	else:
		resultant_status,data=output
	for i,row in control_desk.iterrows():
		if func==row.gen_func:
			if resultant_status==row.ret_stat or (type(row.ret_stat) is types.FunctionType and row.ret_stat(resultant_status)):
				logsup.logger.info('resultant_status: %s,data: %s',resultant_status,data)
				queue.put((row.name,data))

def queue_processor(queue: mp.Queue,control_desk):
	# simple_logger.info('control_desk: %s',control_desk)
	row_name,data=queue.get()
	logsup.logger.pop()
	try:
		func=control_desk.loc[row_name].doer_func
		output=func(data)
		if output is None: return
		queue_generator(queue,func,control_desk,output)
	
	except Exception as e:
		logsup.logger.error(str(e),exc_info=True)
		logsup.logger.inform(str(e),exc_info=True)
		ScheduledTask.on_process_error(data,e)
		raise
	pass

class Conveyor:
	
	def __init__(self,max_queue_size=1,name=''):
		self.control_desk=pd.DataFrame(columns=['gen_func','ret_stat','doer_func'])
		self.queue=mp.Queue()
		self.max_q_size=max_queue_size
		self.positions: List[ConveyorPosition]=list()
		self.funcs_to_run_at_start=set()
		self.has_doer=False
		self.name=name or uuid.uuid4()
	
	def add_control_row(self,*args):
		self.control_desk.loc[self.control_desk.shape[0]]=args
	
	def add_data_provider(self,scheduled_job: schedule.Job,func,run_at_start=True):
		if run_at_start: self.funcs_to_run_at_start.add(func)
		self.add_control_row(func,True,None)
		scheduled_job.do(queue_generator,self.queue,func,self.control_desk).tag(self.name)
		return self
	
	def add_return_status_info(self,some_func,returns_status=True):
		for i,row in self.control_desk.iterrows():
			if some_func==row.gen_func:
				if row.ret_stat is True:
					row.ret_stat=returns_status
					return
		self.add_control_row(some_func,returns_status,None)
	
	def add_doer_func(self,doer_func,max_workers=1):
		# self.control_desk.loc[self.control_desk['doer_func'].isnull() & (self.control_desk['ret_stat'].notnull()),'doer_func']=doer_func
		check1=(self.control_desk['doer_func'].isnull()&self.control_desk['ret_stat'].notnull())
		if check1.any():
			self.control_desk.loc[check1,'doer_func']=doer_func
		else:
			self.control_desk.loc[self.control_desk['doer_func'].isnull()&
			                      (self.control_desk['ret_stat'].isnull()),
			                      ['ret_stat','doer_func']]=True,doer_func
		# print(self.control_desk)
		# raise NotImplementedError
		# for i,row in self.control_desk.iterrows():
		# 	if row.ret_stat is None:
		# 		self.add_return_status_info(row.gen_func)
		# if row.doer_func is None:
		# 	if row.ret_stat is not None:
		# 		row.doer_func=doer_func
		
		#
		# if self.has_doer:
		# 	if row.doer_func is None and row.ret_stat is None:
		# 		row.doer_func=doer_func
		# 		row.ret_stat=True
		# else:
		# 	if row.doer_func is None and row.ret_stat is not None:
		# 		row.doer_func=doer_func
		self.__add_position(doer_func,max_workers=max_workers)
		self.has_doer=True
	
	def __add_position(self,worker_func,max_workers=1):
		conv_position=ConveyorPosition(job_func=worker_func,max_process_count=max_workers)
		self.positions.append(conv_position)
	
	def run_start_funcs(self):
		for func in self.funcs_to_run_at_start:
			queue_generator(self.queue,func,self.control_desk)
	
	def run_pending(self):
		# simple_logger.info('conveyor: %s, qsize: %s',self.name,self.queue.qsize())
		for conv_pos in self.positions:
			conv_pos.check_for_completed_processes()
			conv_pos.start_process(queue_processor,self.queue,self.control_desk)
		if self.queue.qsize()<self.max_q_size:
			schedule.run_pending(self.name)
		pass
# endregion

class ScheduledTask:
	
	def __init__(self,parent_task=None,filename=None,log_timing_opts=None,**setup_kwargs):
		self.parent_task: ScheduledTask=parent_task
		self.filename=filename  # or self.parent_task.filename
		if not self.filename is None:
			# self.create_log_file(level=logsup.INFORM,**(log_timing_opts or {}))
			logsup.logger.create(self.filename,self.__class__.__name__)\
				.get_filepath('main',include_depth=1,include_datetime=2)\
				.add_timed_handler(level=logsup.INFORM,**(log_timing_opts or {}),
			                       log_format=logsup.simple_fmt_no_level)
		else: self.filename=self.parent_task.filename
		# super(scheduled_task,self).__init__(self.__class__.__name__,logsup.log_folder.create(self.__class__.__name__),level=logsup.INFORM,**(log_timing_opts or {}))
		# self.get_filepath('main',include_depth=1,include_datetime=2).add_handler()
		if parent_task is not None: parent_task.child_tasks.append(self)
		self.child_tasks=list()
		self.conveyors: List[Conveyor]=list()
		self.current_conveyor: Conveyor=None
		self.setup(**setup_kwargs)
	
	def setup(self,**kwargs):
		pass
	
	# def create_log_file(self,**kwargs):
	# 	assert self.filename is not None
	# 	logsup.logger.create_log_file(self.filename,process=self.__class__.__name__,**kwargs)
	def log_create(self,*child_folders,**info_kwargs):
		return logsup.logger.create(self.filename,self.__class__.__name__,*child_folders,**info_kwargs)
	
	def check_for(self,func,scheduler,check_at_start=True,**kwargs):
		if self.current_conveyor is None or self.current_conveyor.has_doer:
			c=Conveyor(**dict(dict(name=f"Conveyor-{len(self.conveyors)}"),**kwargs))
			self.conveyors.append(c)
			self.current_conveyor=c
		# self.checker_functions.append(func)
		# self.checker_functions[q]=func
		# self.conveyor_queues[self.current_conveyor].append(q)
		self.current_conveyor.add_data_provider(scheduled_job=scheduler,func=func,run_at_start=check_at_start)
		return self
	
	def when(self,some_func=None,return_status=True):
		if some_func is None: some_funcs=self.current_conveyor.control_desk['gen_func'].tolist()
		else: some_funcs=[some_func]
		if self.current_conveyor.has_doer:
			for some_func in some_funcs:
				self.current_conveyor.add_control_row(some_func,return_status,None)
		else:
			for func in some_funcs:
				# assert func in gen_funcs,"The function should be in the current conveyor's list of generator functions"
				self.current_conveyor.add_return_status_info(some_func=func,returns_status=return_status)
		return self
	
	def do(self,func,workers_count=1):
		self.current_conveyor.add_doer_func(func,max_workers=workers_count)
		self.current_conveyor.add_control_row(func,None,None)
		return self
	
	def run(self,*tasks,sleep_time=1):
		all_tasks=[self,*tasks,*self.child_tasks]
		for task in all_tasks:
			for conveyor in task.conveyors: conveyor.run_start_funcs()
		logsup.logger.info('Start of the main loop.')
		while True:
			for task in all_tasks:
				for conveyor in task.conveyors:
					conveyor.run_pending()
			time.sleep(sleep_time)
	
	@staticmethod
	def on_process_error(data,exception):
		logsup.logger.exception("Exception caught when processing data:\n%s\nException text:\n%s",data,str(exception))

def main():
	pass

if __name__=='__main__':
	main()



import glob
import logging
import os
import shutil
import time
import zipfile
from datetime import datetime
from typing import Callable

import numpy as np
import pandas as pd
import itertools
import collections

from box import Box
from checksumdir import dirhash

from lgblkb_tools import log_support as logsup
from lgblkb_tools.locations import get_parent_dir,get_splitted,InfoDict,create_path,_make_zipfile,ZipError,get_existing_path,get_name,CopyError

class Folder(object):
	
	def __init__(self,path='',pseudo=False,parent=None,propagate_type=False,assert_exists=False):
		if isinstance(path,Folder): path=path.path
		assert isinstance(pseudo,bool)
		if not pseudo:
			if not path: path=os.getcwd()
			elif not os.path.exists(path):
				if assert_exists: raise AssertionError(f'The path {path} should exist.',dict(path=path))
				assert os.path.splitext(os.path.split(path)[-1])[-1]=='',f"Please, use folder path as input. Provided path is: \n{path}"
				os.makedirs(path)
			elif not os.path.isdir(path):
				path=get_parent_dir(path,1)
		self.path=path
		self.pseudo=pseudo
		self.__parent=parent
		self.__propagate_type=propagate_type
	
	@classmethod
	def create_for_filepath(cls,filepath):
		os.makedirs(get_parent_dir(filepath),exist_ok=True)
		return cls(filepath)
	
	@property
	def name(self):
		return os.path.split(self.path)[-1]
	
	def get_filepath(self,*name_portions,ext='',delim='_',include_depth=None,datetime_loc_index=None,**name_kwargs):
		parts=list()
		if include_depth not in [0,None]:
			if type(include_depth) is Callable:
				parent_parts=include_depth(get_splitted(self.path))
			elif type(include_depth) is int:
				parent_parts=get_splitted(self.path)[-include_depth:]
			elif type(include_depth) in [tuple,list]:
				parent_parts=[get_splitted(self.path)[-x] for x in include_depth]
			else:
				raise NotImplementedError(f'include_depth={include_depth}')
			parts.extend(parent_parts)
		part_portions=delim.join([str(x) for x in name_portions])
		if part_portions: parts.append(part_portions)
		part_kwargs=InfoDict(name_kwargs).get_portions()
		if part_kwargs: parts.extend(part_kwargs)
		if datetime_loc_index is not None: parts.insert(datetime_loc_index,datetime.now().strftime("%Y%m%d-%H:%M:%S"))
		assert parts,'Nothing if provided to create filepath.'
		return os.path.join(self.path,delim.join(parts).replace(' ',delim)+ext)
	
	def _mkdir(self,child_dirpath):
		return create_path(self.path,child_dirpath)
	
	def create(self,*child_folders,**info_kwargs):
		parts=list()
		if child_folders: parts.extend(child_folders)
		if info_kwargs: parts.extend(InfoDict(info_kwargs).get_portions())
		assert parts,'Nothing is provided to create directory'
		child_dirpath='__'.join(map(str,parts))
		
		if self.pseudo: path=os.path.join(self.path,child_dirpath)
		else: path=self._mkdir(child_dirpath)
		if self.__propagate_type:
			return self.__class__(path,pseudo=self.pseudo,parent=self)
		else:
			return Folder(path,pseudo=self.pseudo,parent=self,propagate_type=True)
	
	def delete(self):
		if self.pseudo: raise OSError('Cannot delete when in "pseudo" mode.')
		shutil.rmtree(self.path)
	
	def clear(self):
		if self.pseudo: raise OSError('Cannot delete when in "pseudo" mode.')
		self.delete()
		create_path(self.path)
	
	def children(self,*paths):
		return glob.glob(os.path.join(self.path,*(paths or '*')))
	
	@logsup.with_logging(log_level=logging.DEBUG)
	def zip(self,zip_filepath='',save_path_formatter=None,forced=False):
		max_attempts=5
		if not zip_filepath: zipfile_basepath=self.parent().get_filepath(self.name)
		elif zip_filepath[-4:]=='.zip': zipfile_basepath=zip_filepath[:-4]
		else: zipfile_basepath=zip_filepath
		if save_path_formatter is not None: zipfile_basepath=save_path_formatter(zipfile_basepath)
		if not forced:
			fullpath=zipfile_basepath+'.zip'
			if os.path.exists(fullpath) and not zipfile.ZipFile(fullpath).testzip():
				return fullpath
		
		for i in range(max_attempts):
			fullpath=_make_zipfile(base_name=zipfile_basepath,root_dir=get_parent_dir(self.path),base_dir=self.name)
			#shutil.make_archive(base_name=zipfile_basepath,format='zip',root_dir=get_parent_dir(self.path),base_dir=self.name)
			zip_obj=zipfile.ZipFile(fullpath)
			if not zip_obj.testzip():
				return fullpath
			else:
				os.remove(fullpath)
		raise ZipError(f'Could not zip {self.path} to {zipfile_basepath}.zip.')
	
	@logsup.with_logging(log_level=logging.DEBUG)
	def zip_to(self,dest_folder,zipname='',save_path_formatter=None,forced=False):
		return self.zip(Folder(dest_folder).get_filepath(zipname or self.name),save_path_formatter=save_path_formatter,forced=forced)
	
	@logsup.with_logging(log_level=logging.DEBUG)
	def unzip(self,zip_filepath,create_subdir=True):
		# if save_path_formatter is None: save_path_formatter=lambda x:x
		zip_path=[zip_filepath,
		          zip_filepath+'.zip',
		          self.get_filepath(zip_filepath),
		          self.get_filepath(zip_filepath,ext='.zip')]
		zip_path=get_existing_path(zip_path)
		assert not os.path.isdir(zip_path),f'The folder "{zip_filepath}" cannot be unzipped.'
		if create_subdir:
			zipfilename=get_name(zip_path)
			# self.get_filepath(zipfilename)
			shutil.unpack_archive(zip_path,self.create(zipfilename).path,'zip')
		
		else:
			shutil.unpack_archive(zip_path,self.path,'zip')
		return self
	
	def glob_search(self,*patterns,recursive=True):
		return glob.glob(self.get_filepath(*patterns),recursive=recursive)
	
	def find_item(self,partial_name,ending='*'):
		filepath=self.glob_search(f'**/*{partial_name}{ending}')
		if filepath: return filepath[0]
		else: return ''
	
	def rename(self,new_name):
		new_path=os.path.join(get_parent_dir(self.path),new_name)
		os.rename(self.path,new_path)
		self.path=new_path
		return self
	
	def move(self,dst_path,create_subdir=True):
		if self.__propagate_type: classtype=self.__class__
		else: classtype=Folder
		if create_subdir: destination_folder=classtype(dst_path).create(self.name)
		else: destination_folder=classtype(dst_path)
		shutil.move(self.path,destination_folder.path)
		return destination_folder
	
	def parent(self,depth=1):
		if self.__propagate_type: classtype=self.__class__
		else: classtype=Folder
		return classtype(get_parent_dir(self.path,depth=depth),pseudo=self.pseudo)
	
	@logsup.with_logging(log_level=logging.DEBUG)
	def copy_to(self,dst_path,create_subdir=True,forced=False):
		# if self.__propagate_type: classtype=self.__class__
		# else: classtype=partial(Folder,pseudo=True)
		if create_subdir: destination_folder=self.__class__(self.__class__(dst_path).create(self.name),
		                                                    pseudo=self.pseudo,propagate_type=self.__propagate_type)
		else: destination_folder=self.__class__(dst_path,pseudo=self.pseudo,propagate_type=self.__propagate_type)
		
		logsup.logger.debug(f'Copying {self.path} to {destination_folder.path}...')
		source_hashsum=dirhash(self.path)
		if not forced and os.path.exists(destination_folder.path):
			destination_hashsum=dirhash(destination_folder.path)
			if source_hashsum==destination_hashsum: return destination_folder
		max_tries=3
		for i in range(3):
			#if os.path.exists(destination_folder.path):
			shutil.rmtree(destination_folder.path,ignore_errors=True)
			shutil.copytree(self.path,destination_folder.path)
			destination_hashsum=dirhash(destination_folder.path)
			if source_hashsum==destination_hashsum: return destination_folder
			else:
				logsup.logger.warning('source_hashsum',source_hashsum)
				logsup.logger.warning('destination_hashsum',destination_hashsum)
				logsup.logger.warning(f'Warning!!!!!!!!!. Copied folder has different hashsum than its source. Try {i}/{max_tries}. Waiting for 10 sec.')
				time.sleep(10)
		raise CopyError(f'{self} could not be copied to {destination_folder}.',dict(source_folder=self.path,
		                                                                            destination_folder=destination_folder.path))
	
	def get_size_in_gb(self):
		total_size=0
		for dirpath,dirnames,filenames in os.walk(self.path):
			for f in filenames:
				fp=os.path.join(dirpath,f)
				total_size+=os.path.getsize(fp)
		return total_size*1e-9
	
	def __getitem__(self,item):
		if os.path.splitext(item)[-1]:
			return self.get_filepath(item)
		else:
			return self.create(item)
	
	def __setitem__(self,key,value):
		filename,ext=os.path.splitext(key)
		if isinstance(value,str): string_obj=value
		elif type(value) in [list,tuple]:
			string_obj='\n'.join(map(str,value))
		elif isinstance(value,dict):
			ext=ext or '.yaml'
			if ext in ['.yaml','.yml']:
				Box(value).to_yaml(filename=self.get_filepath(filename,ext=ext))
				return
			elif ext=='.json':
				Box(value).to_json(filename=self.get_filepath(filename,ext=ext))
				return
			else:
				string_obj='\n'.join(map(lambda kv:f"{kv[0]}: {kv[1]}",value.items()))
		
		else: string_obj=str(value)
		# def writer(filehandler):
		# 	for k,v in value.items():
		# 		filehandler.writelines([f"{k}: {v}"])
		
		with open(self.get_filepath(filename,ext=ext or '.txt'),'w') as fh:
			fh.write(str(string_obj))
	
	def __repr__(self):
		return f"{self.__class__.__name__}(r'{self.path}')"



from geoalchemy2.types import WKBElement
from geoalchemy2.shape import to_shape
from sqlalchemy.sql.functions import Function as sql_function

def get_info(obj):
	if not isinstance(obj,dict): obj=obj.__dict__
	out_info=dict()
	for k,v in obj.items():
		if k[0]=='_': continue
		elif type(v) in [WKBElement,sql_function]: v=to_shape(v).wkt
		out_info[k]=v
	# info={k:v for k,v in self.__dict__.items() if (k[0]!='_' and not isinstance(v,WKBElement))}
	return out_info

def main():
	pass

if __name__=='__main__':
	main()



import os
from box import Box
import collections
import itertools
import json
import geojson
import geojsonio
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
import shapely.geometry as shg
import shapely.ops as shops
import shapely.wkb as shwkb
import shapely.wkt as shwkt
from functools import partial
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from . import log_support as logsup
from .global_support import Folder
# import visvalingamwyatt as vw

geom_kz_coarse=r'POLYGON ((59.501953125 51.39920565355378, 51.50390625 52.50953477032727, 46.40625 50.875311142200765, '\
               r'45.1318359375 48.516604348867475, 51.37207031249999 41.04621681452063, 56.4697265625 40.713955826286046, '\
               r'58.00781249999999 44.49650533109348, 68.466796875 40.04443758460856, 80.595703125 41.60722821271717, '\
               r'85.869140625 46.46813299215554, 89.1650390625 49.009050809382046, 77.2998046875 54.87660665410869, '\
               r'70.09277343749999 56.07203547180089, 60.1171875 54.316523240258256, 59.501953125 51.39920565355378))'
geom_kz=r'POLYGON ((69.5654296875 55.727110085045986, 60.64453125000001 54.213861000644926, '\
        r'60.941162109375 53.126998061776156, 59.64477539062499 51.93749209045435, 61.31469726562501 51.23440735163459, '\
        r'61.22680664062499 50.89610395554359, 60.00732421875 50.965346321637696, 59.58984374999999 50.736455137010665, '\
        r'58.348388671875 51.29627609493991, 56.304931640625 51.16556659836182, 55.73364257812499 50.72254683363231, '\
        r'54.64599609375 51.193115244645874, 52.33886718749999 51.84935276370605, 50.60302734375 51.86970795368951, '\
        r'48.394775390625 50.680797145321655, 48.69140625 50.05713877598689, 48.372802734375 49.930008124606886, '\
        r'47.63671875 50.56230444080573, 47.13134765625 50.324422739309384, 46.7578125 49.83798245308484, '\
        r'46.318359375 48.38544219115483, 47.21923828125 47.56911375866714, 47.98828124999999 47.628380027447136, '\
        r'48.603515625 46.965259400349275, 48.240966796875 46.63435070293566, 48.49365234375 42.293564192170095, '\
        r'52.49267578125 41.50034959128928, 54.11865234375 42.187829010590825, 55.48095703125 41.08763212467916, '\
        r'56.129150390625 41.244772343082076, 56.085205078125 44.933696389694674, 58.502197265625 45.48324350868221, '\
        r'59.80957031249999 44.933696389694674, 60.8642578125 44.32384807250689, 61.89697265624999 43.42100882994726, '\
        r'63.24829101562501 43.5326204268101, 64.92919921875 43.476840397778936, 65.775146484375 42.71473218539458, '\
        r'65.89599609375 41.86137915587359, 66.324462890625 41.763117447005875, 66.62109375 41.054501963290505, '\
        r'68.73046875 40.421860362045194, 68.807373046875 40.88029480552824, 71.180419921875 42.23665188032057, '\
        r'71.47705078125 42.62587560259137, 73.63037109375 42.309815415686664, 73.795166015625 42.94033923363181, '\
        r'75.684814453125 42.64204079304426, 77.6513671875 42.70665956351041, 78.892822265625 42.62587560259137, '\
        r'80.35400390625 42.00032514831621, 81.10107421874999 43.34914966389313, 80.43090820312499 44.91035917458495, '\
        r'82.880859375 45.1742925240767, 82.562255859375 45.598665689820635, 83.21044921875 46.965259400349275, '\
        r'85.14404296875 46.717268685073954, 85.95703125 47.15236927446393, 85.924072265625 48.20271028869972, '\
        r'86.748046875 48.545705491847464, 87.51708984375 49.33944093715546, 86.759033203125 50.035973672195496, '\
        r'85.58349609375 49.76707407366792, 83.56201171875 51.213765926935025, 81.947021484375 50.90303283111257, '\
        r'80.74951171875 51.50190410761811, 80.1123046875 51.01375465718821, 78.057861328125 53.38332836757156, '\
        r'76.86035156249999 53.97547425742953, 77.1240234375 54.51470449573694, 76.695556640625 54.55295156056188, '\
        r'73.992919921875 53.74221377343122, 73.883056640625 54.194583360162646, 72.158203125 54.508326500290735, '\
        r'71.466064453125 54.322931143263474, 71.378173828125 54.87028529268185, 70.916748046875 55.416543608580064, '\
        r'69.5654296875 55.727110085045986))'

# @with_logging()
def remove_redundants(polygons,redundancy_order=2.5,**kwargs):
	area_max=max([g.area for g in polygons])
	candidates=list()
	for g in list(polygons):
		if g.area==0 or np.log10(area_max/g.area)>redundancy_order:
			logsup.logger.debug('Redundant land is detected. Removing it...')
			continue
		g_ext=g.exterior
		# simple_logger.debug('g_ext:\n%s',g_ext)
		# xy=np.stack([np.array(arr) for arr in g_ext.xy]).T
		# xy=vw.simplify(xy,ratio=0.95)
		# g_ext=shg.LinearRing(xy)
		# simple_logger.debug('xy:\n%s',xy)
		# simple_logger.debug('xy:\n%s',xy)
		# x=np.array(x)
		# simple_logger.debug('x:\n%s',x)
		# simple_logger.debug('y:\n%s',y)
		# raise NotImplementedError
		# simple_logger.debug('xy:\n%s',xy)
		
		g_interiors=list()
		for g_interior in g.interiors:
			gint=shg.Polygon(g_interior)
			if gint.area==0 or np.log10(area_max/gint.area)>redundancy_order:
				logsup.logger.debug('Redundant hole is detected. Removing it...')
				continue
			g_interiors.append(g_interior)
		clean_g=shg.Polygon(g_ext,g_interiors)
		candidates.append(clean_g)
	
	return candidates

def get_crs(geojson_datum):
	crs_info=geojson_datum.crs.properties['name'].split(':')
	if not crs_info[-1].isnumeric():
		print(crs_info)
		raise NotImplementedError('Could not infer crs. crs_info = \n{}'.format(crs_info))
	return int(crs_info[-1])

def plot_polygon(polygon: shg.Polygon,label=None,ax=None,name='',**kwargs):
	if ax is None: plot=plt.plot
	else: plot=ax.plot
	default_opts=dict(lw=3)
	if label is None: labeler=lambda l:dict(default_opts,label=l)
	else: labeler=lambda l:dict(default_opts,label='{} {}'.format(label,l))
	plotter=lambda linestring,_label:plot(*linestring.xy,**dict(labeler(_label),**kwargs))
	plotter(polygon.exterior,'exterior')
	if name:
		x,y=polygon.exterior.xy
		target_point_index=list(x).index(min(x))
		
		plt.text(x[target_point_index],y[target_point_index],s=name)
	for i,p_i in enumerate(polygon.interiors):
		plotter(p_i,'interior {}'.format(i))

def plot_poly(p,**kwargs):
	if isinstance(p,shg.MultiPolygon):
		for p in p: plot_poly(p,**kwargs)
	plt.plot(*p.exterior.xy,**kwargs)

def convert_crs(coor_obj,_in: int,_out=4326):
	inproj,outproj=pyproj.Proj(init=f'epsg:{_in}'),pyproj.Proj(init=f'epsg:{_out}')
	converter=lambda xy:np.stack(pyproj.transform(inproj,outproj,*xy),axis=1).reshape(-1,2)
	polygons=list()
	if type(coor_obj) is np.ndarray:
		assert coor_obj.shape[1]==2,\
			"ndarray shape should be (-1,2), but shape={} is provided.".format(coor_obj.shape)
		return converter(np.split(coor_obj,2,axis=1))
	elif type(coor_obj) is shg.Polygon:
		coor_obj=shg.MultiPolygon([coor_obj])
		returner=lambda multipolygon:multipolygon[0]
	elif type(coor_obj) is shg.MultiPolygon:
		returner=lambda multipolygon:multipolygon
		pass
	else: raise NotImplementedError('No implementation for objects of type = {}'.format(type(coor_obj)))
	for g in coor_obj:
		interior_xys=[converter(interior.xy) for interior in g.interiors]
		polygons.append(shg.Polygon(converter(g.exterior.xy),interior_xys))
	return returner(shg.MultiPolygon(polygons))

def reproject_point(point,_in,_out):
	inproj,outproj=pyproj.Proj(init=f'epsg:{_in}'),pyproj.Proj(init=f'epsg:{_out}')
	return shg.Point(pyproj.transform(inproj,outproj,point.x,point.y))

def get_area(geom,in_epsg):
	geom_area=shops.transform(
		partial(
			pyproj.transform,
			pyproj.Proj(init=f'EPSG:{in_epsg}'),
			pyproj.Proj(proj='aea',lat1=geom.bounds[1],lat2=geom.bounds[3])),geom).area
	return geom_area

def clean_multipolygon(multipolygon: shg.MultiPolygon):
	geoms=list()
	tree=collections.defaultdict(list)
	# multipolygon=multipolygon.buffer(1e-3)
	for polygon in multipolygon:
		exterior_geom=shg.Polygon(polygon.exterior)
		geoms.append(exterior_geom)
		geoms.extend([shg.Polygon(x) for x in polygon.interiors])
	for g1,g2 in itertools.combinations(geoms,2):
		if g1.contains(g2):
			#
			# if tree.get(g1):
			# 	g1_obj=tree.get(g1)
			# 	if g1_obj.contains(g2):
			# 		tree[g1]=g1_obj.difference(g2)
			# 	pass
			# else:
			# 	tree[g1]=g1.difference(g2)
			tree[geoms.index(g1)].append(g2)
	
	clean_polygons=list()
	for shell_index,holes in tree.items():
		try:
			p=shg.Polygon(shg.LinearRing(geoms[shell_index].exterior.coords),
			              [shg.LinearRing(x.exterior.coords) for x in holes])
		except IndexError as exc:
			logsup.logger.debug('multipolygon:\n%s',multipolygon)
			logsup.logger.debug('geoms:\n%s',geoms)
			logsup.logger.debug('shell_index: %s',shell_index)
			raise exc
		geoms.pop(shell_index)
		
		for x in holes: geoms.remove(x)
		clean_polygons.append(p)
	
	# p=shg.Polygon(shg.LinearRing(geoms[shell_index].coords.xy),[shg.LinearRing(x.coords.xy) for x in holes])
	# clean_polygons.append(p)
	
	# result=g1.difference(g2)
	# simple_logger.info('result: %s',result)
	# simple_logger.debug('tree:\n%s',tree)
	
	return shg.MultiPolygon([*clean_polygons,*geoms])

class SpatialGeom:
	__protected_members=['name','data','geom_obj','box']
	
	@classmethod
	def from_geojson(cls,geojson_datum,crs_in=None,crs_out=4326,**kwargs):
		if isinstance(geojson_datum,str):
			geojson_datum=geojson.load(open(geojson_datum))
		if 'features' in geojson_datum.keys(): geometry=geojson_datum.features[0].geometry
		else: geometry=geojson_datum.geometry
		cad_land=SpatialGeom(shg.shape(geometry),**kwargs)
		if cad_land.geom_obj.is_empty:
			print('Empty entry detected.')
			return cad_land
		cad_land.data['epsg']=crs_out
		if not 'name' in kwargs: cad_land.name=geojson_datum['features'][0]['properties']['KAD_NOMER']
		if crs_in==crs_out: return cad_land
		elif crs_in is None:
			if 'crs' in geojson_datum.keys():
				crs_in=get_crs(geojson_datum)
				if crs_in==crs_out: return cad_land
			elif max(cad_land.geom_obj.bounds)>90:
				#print('Assuming input crs = epsg: 32643')
				#crs_in='epsg:32643'
				raise NotImplementedError(
					'No info to infer from. Please, specify incoming coordinate reference system.')
			else:
				#print('Assuming proper input crs, i.e. "epsg:4326".')
				#return cad_land
				raise NotImplementedError(
					'No info to infer from. Please, specify incoming coordinate reference system.')
		cad_land.convert_crs(crs_in,crs_out)
		return cad_land
	
	def convert_crs(self,_from: int,to=4326):
		if self.geom_obj.is_empty:
			print('Cannot convert empty shape.')
			return self
		if _from==to: return self
		multi_polygon=convert_crs(self.geom_obj,_from,to)
		#if inplace:
		self.geom_obj=multi_polygon
		return self
		# else:
		# 	return SpatialGeom(multi_polygon,redundancy_order=self.redundancy_order,
		# 	                   name=self.name)
		pass
	
	def __init__(self,geom_obj,name='',**kwargs):
		"""

		:param geom_obj: wkt/wkb string, ndarray or shapely Polygon/Multipolygon
		:param name: name
		:param kwargs:
		redundancy_order = 2.5
		"""
		if isinstance(geom_obj,str):
			try:
				geom_obj=shwkt.loads(geom_obj)
			except Exception as e:
				try:
					json_obj=geojson.loads(geom_obj)
					geom_obj=shg.shape(json_obj['features'][0]['geometry'])
				except Exception as e:
					from shapely import geos
					# geos.WKBWriter.defaults['include_srid']=True
					geom_obj=shwkb.loads(geom_obj,hex=True)
		
		if isinstance(geom_obj,shg.Polygon): geom_obj=shg.MultiPolygon([geom_obj])
		elif isinstance(geom_obj,np.ndarray):
			geom_obj=shg.MultiPolygon([shg.Polygon(geom_obj)])
		self.name=name# or 'SpatialGeom'
		self.data=Box(kwargs)
		if not geom_obj.is_empty: geom_obj=shg.MultiPolygon(remove_redundants(geom_obj,**kwargs))
		self.geom_obj: shg.MultiPolygon=clean_multipolygon(geom_obj)
		self.data['with_holes']=max(self.do_for_each(lambda p:len(list(p.interiors))),default=0)!=0
		pass
	
	def do_for_each(self,func,*args,**kwargs):
		doer=lambda x:func(x,*args,**kwargs)
		#if self.is_multipolygon:
		out=list()
		for p in self.geom_obj:
			out.append(doer(p))
		return out
	
	# else:
	# 	return [doer(self.geom_obj)]
	
	def show_on_geojsonio(self):
		get_displayed=lambda obj:geojsonio.display(json.dumps(shg.mapping(obj)))
		self.do_for_each(get_displayed)
	
	def plot(self,show=False,**kwargs):
		if self.geom_obj.is_empty: return
		# if ax is None: plot=plt.plot
		# else: plot=ax.plot
		# plotter=lambda linearring,label,:plot(*linearring.xy,**dict(dict(label=label),**kwargs))
		#if self.is_multipolygon:
		for i,p in enumerate(self.geom_obj):
			plot_polygon(p,**dict(dict(label='Polygon {}'.format(i),name=self.name),**kwargs))
		if show:
			plt.legend(loc='best')
			plt.show()
		return self
	
	def to_geojson(self,**kwargs):
		return geojson.Feature(geometry=self.geom_obj,**kwargs)
	
	def __bool__(self):
		return self.geom_obj.is_empty
	
	# else:
	# 	plot_polygon(self.geom_obj,'Single polygon',**kwargs)
	
	def get_area(self,in_epsg=4326):
		return get_area(self.geom_obj,in_epsg)
	
	def intersection(self,geom,name=None):
		if isinstance(geom,SpatialGeom): geom=geom.geom_obj
		return SpatialGeom(self.geom_obj.intersection(geom),name=name)
	
	def difference(self,geom,name=None):
		if isinstance(geom,SpatialGeom): geom=geom.geom_obj
		return SpatialGeom(self.geom_obj.difference(geom),name=name)
	
	def __getitem__(self,item):
		return self.data[item]
	
	def __iter__(self):
		return iter(self.data.keys())
	
	def __getattr__(self,item):
		if item in self.__protected_members: return self.__getattribute__(item)
		else: return self.data[item]
	
	@property
	def bounding_box(self):
		return SpatialGeom(shg.Polygon(self.geom_obj.envelope.exterior))
	
	def save(self,folder,filename=None,filepath=None):
		if filepath is None: filepath=Folder(folder).get_filepath(filename or self.name or 'SpatialGeom')
		if not os.path.splitext(filepath)[-1]: filepath=filepath+'.svg'
		plt.savefig(filepath,dpi=None,facecolor='w',edgecolor='w',
		            orientation='portrait',papertype=None,format=None,
		            transparent=False,bbox_inches=None,pad_inches=0.1,
		            frameon=None,metadata=None)
		plt.clf()
		return self
	
	def __str__(self):
		return '{}: {}'.format(self.name,str(self.geom_obj))
	
	def buffer(self,distance,resolution=16,quadsegs=None,cap_style=shg.CAP_STYLE.round,
	           join_style=shg.JOIN_STYLE.round,mitre_limit=5.0):
		return self.__class__(self.geom_obj.buffer(distance,resolution=resolution,quadsegs=quadsegs,
		                                           cap_style=cap_style,join_style=join_style,mitre_limit=mitre_limit))
#
# @property
# def holes(self):
# 	if self.__holes is None:
# 		self.__holes=[line_xy(x) for x in self.polygon.boundary[1:]]
# 	return self.__holes
#
# @property
# def all_xy(self):
# 	return np.concatenate([self.xy[:-1],*[x[:-1] for x in self.holes]])
#
# @property
# def xy(self):
# 	if self.__xy is None:
# 		if self.is_multilinestring:
# 			self.__xy=line_xy(self.polygon.boundary[0])
# 		else:
# 			self.__xy=line_xy(self.polygon.boundary)
# 	return self.__xy

def plot_patches(polygons,ax=None,c=None,**kwargs):
	if ax is None: ax=plt.gca()
	if isinstance(polygons,pd.Series):
		pcol=PatchCollection(polygons.map(lambda p:Polygon(p.exterior)),**kwargs)
	else:
		poly_objs=[Polygon(x.exterior) for x in polygons]
		pcol=PatchCollection(poly_objs,**kwargs)
	if c: pcol.set_color(c)
	ax.add_collection(pcol)

def get_rect(corner,sizes):
	corner=np.array(corner)
	sizes=np.array(sizes)
	return np.array([corner,corner+sizes*(1,0),corner+sizes,corner+sizes*(0,1)])

def get_portion(polygon,offset_ratio,size_ratio):
	p_min=np.array(polygon.bounds[0:2])
	p_max=np.array(polygon.bounds[-2:])
	curr_sizes=(p_max-p_min)
	new_rect=shg.Polygon(get_rect(p_min+curr_sizes*offset_ratio,curr_sizes*size_ratio))
	return new_rect

def get_forecast_region(cad_geom,min_forecast_point_dist=3,forecast_step_degrees=0.25):
	forecast_region=cad_geom.envelope.buffer(min_forecast_point_dist*forecast_step_degrees*1.2)
	return forecast_region.envelope

if __name__=='__main__':
	# p=shg.Point(78,43)
	# print(reproject_point(p,4326,3857))
	# print(p.x)
	pass

# a1=shg.Polygon(np.random.rand(4,2))
# a2=shg.Polygon()
# b1=shg.MultiPolygon()
# b2=shg.MultiPolygon([a1])
# b3=shg.MultiPolygon([a1,shg.Polygon(np.random.rand(4,2),[np.random.rand(3,2)])])
# c1=SpatialGeom(a1)
# #c2=c1.convert_crs(32642)
# print(convert_crs(np.random.rand(10,2),32642))
